// SPDX-FileCopyrightText: 2023 Andrei Zisu <matzipan@gmail.com>
// SPDX-FileCopyrightText: 2023 Helge Eichhorn <git@helgeeichhorn.de>
// SPDX-FileCopyrightText: 2024 Angus Morrison <github@angus-morrison.com>
//
// SPDX-License-Identifier: MPL-2.0

//! Mod `types` exposes shared type aliases and utility functions for working with Julian dates.

pub type Epoch = f64;

pub type ModifiedJulianDate = f64;

pub type ModifiedJulianDayNumber = i32;
