// SPDX-FileCopyrightText: 2024 Angus Morrison <github@angus-morrison.com>
// SPDX-FileCopyrightText: 2024 Helge Eichhorn <git@helgeeichhorn.de>
//
// SPDX-License-Identifier: MPL-2.0

//! Module `python` aggregates the Python binding for `lox-time`.

pub mod deltas;
pub mod time;
pub mod time_scales;
pub mod utc;
