// SPDX-FileCopyrightText: 2023 Andrei Zisu <matzipan@gmail.com>
// SPDX-FileCopyrightText: 2023 Helge Eichhorn <git@helgeeichhorn.de>
// SPDX-FileCopyrightText: 2024 Angus Morrison <github@angus-morrison.com>
//
// SPDX-License-Identifier: MPL-2.0

//! Mod `units` exposes shared type aliases representing a variety of physical units.

pub type Arcseconds = f64;

pub type Microarcseconds = f64;

pub type Radians = f64;

pub type Seconds = f64;

pub type JulianCenturies = f64;

pub type Days = f64;
